package massim.agent.framework;

import java.util.concurrent.PriorityBlockingQueue;
import massim.agent.implementation.AgentLauncher;

/**
 * Abstract Dama Synchronous Agent
 */
public abstract class AbstractDamaSyncAgent extends AbstractAsyncAgent {

    /**
     * Inbox with sorted messages according to message timestamp
     */
    private PriorityBlockingQueue<MessageContent> inbox;
    private Perception perception;
    private String action;
    private Boolean bActed = false;
    protected Integer priority = -1;
    protected Integer currentX = -1;
    protected Integer currentY = -1;
    private Boolean bLogging = true;

    /**
     * Constructor for new agent
     * 
     * @param username User name
     * @param password Password
     * @param priority Priority (unique number ID)
     */
    public AbstractDamaSyncAgent(String username, String password, Integer priority) {
        super(username, password);
        this.inbox = new PriorityBlockingQueue<MessageContent>();
        this.perception = null;
        this.priority = priority;
    }

    @Override
    protected final String processRequestAction(Perception perception) {
        synchronized (this) {
            this.perception = perception;
            if (perception != null) {
                currentX = perception.posX;
                currentY = perception.posY;
            }
            action = "skip";
            bActed = false;
        }

        // wait some time for the agent to act
        // this has to remain unlocked in order to allow the agent to reason and act
        // we wait until the agent calles act(), but at most 4*450ms=1800ms
        String returnAction = "skip";
        try {
            Boolean acted = false;
            int wakeUps = 0;
            while ((!acted) && (wakeUps <= 4)) {
                wakeUps++;
                Thread.sleep(450);

                // read the acted flag with caution in a synchronized mode!
                synchronized (this) {
                    acted = bActed;
                    returnAction = action;
                }
            }
        } catch (InterruptedException e) {
            e.printStackTrace();
        }

        // reset action
        synchronized (this) {
            action = "skip";
        }

        // return the action which was set by the agent
        return returnAction;
    }

    /**
     * Returns true if the communication inbox is empty.
     * 
     * The method is non-blocking. I.e., it returns immediately regardless
     * of whether there are messages in the message queue, or not. 
     */
    public synchronized final Boolean isEmptySensor() {
        return (perception == null);
    }

    /**
     * Reset sensor
     */
    public synchronized final void resetSensor() {
        perception = null;
    }

    /**
     *  Message arrival handler for the AsynchronousAgent class
     */
    @Override
    protected synchronized final void onMessageReceivedPrivate(MessageContent content) {
        onMessageReceivedToInbox(content);
    }

    /**
     * The handler of message arrival to the inbox
     * 
     * @param message Arriving message
     */
    private synchronized void onMessageReceivedToInbox(MessageContent message) {
        try {
            // put message in the inbox
            inbox.put(message);
        } catch (Exception e) {
            System.err.println("ERROR: Inbox handling corrupted!");
            e.printStackTrace();
        }
    }

    /**
     * Retrieves a message from the communication inbox.
     * 
     * Watch out! The method is blocking, i.e., in the case there is no message in the inbox
     * it waits for a message to arrive.
     */
    public synchronized final MessageContent popInboxMessage() {
        try {
            // debug
            // System.out.println("Messages in inbox = " + inbox.size());
            return inbox.take(); // take message from inbox
        } catch (InterruptedException e) {
            System.err.println("ERROR: message queue popInboxMessage() was interrupted!");
            e.printStackTrace();
        }

        // something very bad happened!
        return null;
    }

    /**
     * Returns true if the communication inbox is empty.
     * 
     * The method is non-blocking. I.e., it returns immediately regardless of whether
     * there are messages in the message queue, or not. 
     */
    public synchronized final Boolean isEmptyInbox() {
        return inbox.isEmpty();
    }

    /**
     * Handler of inbox message receipt. 
     */
    @Override
    protected synchronized final void onMessageReceived(MessageContent content) {
        System.err.println("Wrong use of AbstractSyncAgent class! Synchronous agent must actively poll its message queue in order to retrieve messages");
    }

    /**
     * Do an action
     * 
     * @param action Action
     */
    private synchronized void act(String action) {
        this.action = action;
        this.bActed = true;
    }

    /**
     * Go left
     */
    public final void actLeft() {
        act("west");
    }

    /**
     * Go right
     */
    public final void actRight() {
        act("east");
    }

    /**
     * Do nothing
     */
    public final void actNothing() {
        act("skip");
    }

    /**
     * The launch method for the synchronous agent implementation. 
     * 
     * Here the user can implement whatever continuous reasoning algorithms 
     * which then interacts with the simulation only when it wants and is 
     * never explicitely bothered by anybody by callbacks. 
     */
    protected abstract void runSyncAgent();

    /**
     * Broadcasting is not possible
     */
    @Override
    public final void broadcastMessage(MessageContent content) {
    }

    /**
     * Send message is not possible - replaced by handleOK, handleNoGood
     */
    @Override
    public final void sendMessage(MessageContent content, String receiver) {
    }

    /**
     * Returns the priority of the agent.
     * 
     * @return The priority
     */
    public Integer getPriority() {
        return priority;
    }

    /**
     * Log communication
     * @param receiver Receiver of the message
     * @param contents Message contents
     */
    private void logMessage(Integer receiver, String contents) {
        if (bLogging) {
            System.out.format("LOG MESSAGE sender=%d receiver=%d step=#%d MSG='%s'\n", priority, receiver, getSimStep(), contents);
        }

    }

    /**
     * Send HANDLE-OK? message to one receiver
     * 
     * Message format: 'HandleOK:priorityOfSender:senderCurrentX,senderCurrentY'
     *                  e.g.: 'HandleOK:2:6,2' 
     * 
     * @param receiver Priority (number ID) of the receiver
     */
    public void sendHandleOK(Integer receiver) {
        String s = String.format("HandleOK:%d:%d,%d", priority, currentX, currentY);
        long ct = System.currentTimeMillis();
        MessageContent msg = new StringContent(s, ct);
        super.sendMessage(msg, AgentLauncher.ID2username(receiver));
        logMessage(receiver, s);
    }

    /**
     * Send HANDLE-NOGOOD to one receiver
     * 
     * Message format: 'HandleNoGood:priorityOfSender:representationOfNoGood'
     *                  e.g.: 'HandleNoGood:2:nogood'
     * 
     * @param receiver Priority (number ID) of the receiver
     * @param nogood String representation of a NOGOOD
     */
    public void sendHandleNoGood(Integer receiver, String nogood) {
        String s = String.format("HandleNoGood:%d:%s", priority, nogood);
        long ct = System.currentTimeMillis();
        MessageContent msg = new StringContent(s, ct);
        super.sendMessage(msg, AgentLauncher.ID2username(receiver));
        logMessage(receiver, s);
    }

    /**
     * Does nothing
     * (perception is available in variables currentX, currentY)
     */
    public final Perception sense() {
        return null;
    }

    // The main entry point into the agent deliberation loop
    @Override
    public void run() {

        // Launch the agent thread
        new Thread(
                new Runnable() {

                    @Override
                    public void run() {
                        runSyncAgent();
                    }
                }).start();

        // You can configure the agent's connection point
        run("localhost", 12300);
    }

    /**
     * Thread sleep
     * 
     * @param ms Number of milliseconds
     */
    protected void sleep(int ms) {
        try {
            Thread.sleep(ms);
        } catch (InterruptedException ex) {
            ex.printStackTrace();
        }
    }
}
